/* File: devMM4000.c 		   	*/
/* Version: 1.01		   	*/
/* Date Last Modified: 10/25/98	        */

/* Device Support Routines for motor */
/*
 *      Original Author: Mark Rivers
 *      Date: 10/16/97
 *
 *      Experimental Physics and Industrial Control System (EPICS)
 *
 *      Copyright 1991, the Regents of the University of California,
 *      and the University of Chicago Board of Governors.
 *
 *      This software was produced under  U.S. Government contracts:
 *      (W-7405-ENG-36) at the Los Alamos National Laboratory,
 *      and (W-31-109-ENG-38) at Argonne National Laboratory.
 *
 *      Initial development by:
 *	      The Controls and Automation Group (AT-8)
 *	      Ground Test Accelerator
 *	      Accelerator Technology Division
 *	      Los Alamos National Laboratory
 *
 *      Co-developed with
 *	      The Controls and Computing Group
 *	      Accelerator Systems Division
 *	      Advanced Photon Source
 *	      Argonne National Laboratory
 *
 * Modification Log:
 * -----------------
 * .00  10-16-97	mlr     initialized from devOms58
 * .01  5-13-98	        mlr     Fixed error in ptrans->offset initialization
 * .02  10-25-98        mlr     Minor code cleanup, fixed formatting
 *      ...
 */


#define VERSION 1.01

#include	<vxWorks.h>
#include        <stdlib.h>
#include	<vme.h>
#include	<types.h>
#include	<stdioLib.h>
#include	<string.h>
#include	<math.h>
#include        <semLib.h>	/* jps: include for init_record wait */
#include	<iv.h>

#include	<alarm.h>
#include	<callback.h>
#include	<dbDefs.h>
#include	<dbAccess.h>
#include	<dbCommon.h>
#include	<fast_lock.h>
#include	<recSup.h>
#include	<devSup.h>
#include	<drvSup.h>
#include	<dbScan.h>
#include	<special.h>
#include	<module_types.h>
#include	<eventRecord.h>
#include	<motorRecord.h>

#include	"motor.h"
#include	"drvMM4000.h"

#define STATIC static

#ifdef NODEBUG
#define Debug(L,FMT,V) ;
#else
#define Debug(L,FMT,V) {  if(L <= devMM4000debug) \
			{ printf("%s(%d):",__FILE__,__LINE__); \
			  printf(FMT,V); } }
#endif

#define NINT(f)	(long)((f)>0 ? (f)+0.5 : (f)-0.5)	/* tmm */
#define ABS(f) ((f)>0 ? (f) : -(f))
#define SIGN(f) (long)((f)>0 ? 1 : -1)
#define MAX(a,b) ((a)>(b) ? (a) : (b))
#define MIN(a,b) ((a)<(b) ? (a) : (b))


/* ----------------Create the dsets for devMM4000----------------- */
/* static long report(); */
STATIC long MM4000_init(int after);
STATIC long MM4000_init_record(struct motorRecord * mr);
STATIC long MM4000_start_trans();
STATIC long MM4000_update_values();
STATIC long MM4000_build_trans();
STATIC long MM4000_end_trans();
STATIC void MM4000_callback();


typedef struct
{
    long number;
    DEVSUPFUN report;
    DEVSUPFUN init;
    DEVSUPFUN init_record;
    DEVSUPFUN get_ioint_info;
    DEVSUPFUN update_values;
    DEVSUPFUN start_trans;
    DEVSUPFUN build_trans;
    DEVSUPFUN end_trans;
} MM4000DSET;

MM4000DSET devMM4000 = {8,
    NULL,
    MM4000_init,
    MM4000_init_record,
    NULL,
    MM4000_update_values,
    MM4000_start_trans,
    MM4000_build_trans,
    MM4000_end_trans
};


/* --------------------------- program data --------------------- */

/* This table is used to define the command types */

static int MM4000_table[] = {
    MOTION, /* MOVE_ABS */
    MOTION, /* MOVE_REL */
    MOTION, /* HOME_FOR */
    MOTION, /* HOME_REV */
    MOTION, /* HOME_ENC */
    IMMEDIATE, /* LOAD_POS */
    IMMEDIATE, /* SET_VELO_BASE */
    IMMEDIATE, /* SET_VELO */
    IMMEDIATE, /* SET_ACCEL */
    IMMEDIATE, /* GO */
    IMMEDIATE, /* SET_ENC_RATIO */
    INFO, /* GET_MOTOR_POS */
    INFO, /* GET_ENCODER_POS */
    INFO, /* GET_INFO */
    MOVE_TERM, /* STOP_AXIS */
    VELOCITY, /* JOG */
};

volatile int devMM4000debug = 0;

/* status of an axis of a board, one present for each axis of each board*/
struct a_axis
{
    int in_use;
};

/* the device private data structure of the record */
struct MM4000_trans
{
    int state;
    FAST_LOCK lock;
    MOTOR_CALL motor_call;
    int callback_changed;
    double motor_pos;
    double encoder_pos;
    double vel;
    double offset;  /* Defined as MM4000_position - mr->rmp*drive_resolution */
    double drive_resolution;
    double high_limit;
    double low_limit;
    unsigned long status;
    SEM_ID initSem;
};

/* the board status structure */
struct MM4000_stat
{
    int exists;
    int total_axis;
    struct a_axis *MM4000_axis;
};

static struct MM4000_stat *MM4000_cards;
extern struct MM4000_support MM4000_access;

#define IDLE_STATE 0
#define BUILD_STATE 1


/* --------------------------- program data --------------------- */


/* initialize device support for MM4000 stepper motor */
STATIC long MM4000_init(int after)
{
    MOTOR_CARD_QUERY card_query;
    int i, j;
    int num_cards;

    Debug(1, "MM4000_init: entry...\n", 0);
    if (after)
	return (0);

    /* Initialize MM4000 driver  */
    (*MM4000_access.init)();

    /* allocate space for total cards in system */
    MM4000_cards = (struct MM4000_stat *) malloc(MM4000_NUM_CARDS *
					   sizeof(struct MM4000_stat));

    /* for each card ask driver for num of axis supported and name */
    for (i = 0; i < MM4000_NUM_CARDS; i++)
    {
	(*MM4000_access.get_card_info) (i, &card_query);

	if (card_query.total_axis == 0)
        {
	    MM4000_cards[i].exists = NO;
	    Debug(5, "MM4000_init: controller %d does not exist\n", i);
        }
	else
	{
	    MM4000_cards[i].exists = YES;
	    MM4000_cards[i].total_axis = card_query.total_axis;
	    Debug(5, "MM4000_init: controller %d: \n", i);
	    Debug(5, "    calling malloc(card_query...) total axis =%d\n", 
                            MM4000_cards[i].total_axis);
	    MM4000_cards[i].MM4000_axis = (struct a_axis *) malloc(
			     card_query.total_axis * sizeof(struct a_axis));

	    for (j = 0; j < card_query.total_axis; j++)
	      {
		MM4000_cards[i].MM4000_axis[j].in_use = 0;
	      }
	}
    }

    Debug(1, "MM4000_init: exit...\n", 0);
    return (0);
}


/* initialize a record instance */
STATIC long MM4000_init_record(struct motorRecord * mr)
{
    int card, signal;
    struct MM4000_trans *ptrans;
    MOTOR_AXIS_QUERY axis_query;
    MOTOR_CALL *motor_call;
    int rtnStat;

    Debug(1, "MM4000_init_record: entry...\n", 0);

    /* allocate space for private field - an MM4000_trans structure */
    Debug(5, "MM4000_init_record: calling malloc(...MM4000_trans...)\n", 0);
    mr->dpvt = (struct MM4000_trans *) malloc(sizeof(struct MM4000_trans));

    ptrans = mr->dpvt;
    ptrans->state = IDLE_STATE;
    ptrans->callback_changed = NO;
    FASTLOCKINIT(&ptrans->lock);
    motor_call = &(ptrans->motor_call);

    callbackSetCallback(MM4000_callback, &(motor_call->callback));
    callbackSetPriority(priorityMedium, &(motor_call->callback));

    /* check if axis already in use, then mark card, axis as in use */

    /* out must be an VME_IO */
    switch (mr->out.type)
    {
    case (VME_IO):
	break;
    default:
	recGblRecordError(S_dev_badBus, (void *) mr,
			  "devMM4000 (init_record) Illegal OUT Bus Type");
	return (S_dev_badBus);
    }

    card = mr->out.value.vmeio.card;
    Debug(5, "MM4000_init_record: card %d\n", card);
    signal = mr->out.value.vmeio.signal;
    Debug(5, "MM4000_init_record: signal %d\n", signal);

    rtnStat = 0;

    if (card < 0 || card >= MM4000_NUM_CARDS || MM4000_cards[card].exists == NO)
    {
	recGblRecordError(S_db_badField, (void *) mr,
			  "devMM4000 (init_record) card does not exist!");
	rtnStat = S_db_badField;
    }
    else if (signal < 0 || signal >= MM4000_cards[card].total_axis)
    {
	recGblRecordError(S_db_badField, (void *) mr,
			  "devMM4000 (init_record) signal does not exist!");
	rtnStat = S_db_badField;
    }
    else if (MM4000_cards[card].MM4000_axis[signal].in_use == YES)
    {
	recGblRecordError(S_db_badField, (void *) mr,
			  "devMM4000 (init_record) motor already in use!");
	rtnStat = S_db_badField;
    }

    if (rtnStat)
    {
        /* Initialize readback fields for simulation */
	mr->msta = RA_PROBLEM;
	mr->res = mr->mres;
	mr->rmp = 0;		/* raw motor pulse count */
	mr->rep = 0;		/* raw encoder pulse count */
	return (rtnStat);
    }
    /* query motor for all info to fill into record */
    (*MM4000_access.get_axis_info) (card, signal, &axis_query);

    MM4000_cards[card].MM4000_axis[signal].in_use = YES;

    mr->rmp = NINT(mr->dval / mr->mres);
    mr->rep = mr->rmp;
    mr->msta = axis_query.status;	/* status info */
    mr->res = mr->mres;
    mr->eres = mr->mres;
    ptrans->drive_resolution = axis_query.drive_resolution;
    ptrans->offset = axis_query.position - mr->rmp*ptrans->drive_resolution; 
    ptrans->high_limit = axis_query.high_limit;
    ptrans->low_limit = axis_query.low_limit;

    Debug(5, "MM4000_init_record: axis_query.position = %f\n", 
                                            axis_query.position);
    Debug(5, "MM4000_init_record: ptrans->offset = %f\n", ptrans->offset);
    Debug(5, "MM4000_init_record: drive_resolution = %f\n",
                                    ptrans->drive_resolution);
    Debug(5, "MM4000_init_record: high_limit = %f\n",
                                    ptrans->high_limit);
    Debug(5, "MM4000_init_record: low_limit = %f\n",
                                    ptrans->low_limit);
    Debug(5, "MM4000_init_record: rmp = %d\n", mr->rmp);
    Debug(5, "MM4000_init_record: rep = %d\n", mr->rep);
    Debug(5, "MM4000_init_record: msta = %d\n", mr->msta);

    Debug(1, "MM4000_init_record: exit...\n", 0);
    return (0);
}


STATIC long MM4000_update_values(struct motorRecord * mr)
{
    struct MM4000_trans *ptrans;
    long rc;

    rc = NOTHING_DONE;
    ptrans = (struct MM4000_trans *) mr->dpvt;

    FASTLOCK(&ptrans->lock);

    /* raw motor pulse count */
    if (ptrans->callback_changed == YES)
    {
	mr->rmp = NINT((ptrans->motor_pos - ptrans->offset) / 
                                ptrans->drive_resolution);
	mr->rep = mr->rmp;
	mr->rvel = NINT(ptrans->vel / ptrans->drive_resolution);
	mr->msta = ptrans->status;
	ptrans->callback_changed = NO;
	rc = CALLBACK_DATA;
    }

    FASTUNLOCK(&ptrans->lock);

    return (rc);
}


/* start building a transaction */
STATIC long MM4000_start_trans(struct motorRecord * mr)
{
    int card = mr->out.value.vmeio.card;
    int axis = mr->out.value.vmeio.signal;
    struct MM4000_trans *trans = (struct MM4000_trans *) mr->dpvt;
    MOTOR_CALL *motor_call;

    if (!mr->dpvt)
	return (S_dev_NoInit);

    motor_call = &(trans->motor_call);

    /*
     * initialize area to device private field to store command that is to be
     * built and mark as command build in progress
     */

    trans->state = BUILD_STATE;
    motor_call->card = card;
    motor_call->signal = axis;
    motor_call->type = UNDEFINED;
    motor_call->precord = (struct dbCommon *) mr;
    motor_call->message[0] = (char) NULL;

    return (0);
}


/* end building a transaction */
STATIC long MM4000_end_trans(struct motorRecord * mr, int term_method)
{
    struct MM4000_trans *trans = (struct MM4000_trans *) mr->dpvt;
    MOTOR_CALL *motor_call;

    motor_call = &(trans->motor_call);

    if (trans->state != BUILD_STATE) return (ERROR);
    /* shut off command build in process thing */
    trans->state = IDLE_STATE;
    return (*MM4000_access.send) (motor_call);
}


/* add a part to the transaction */
STATIC long MM4000_build_trans(int command, double *parms, struct motorRecord * mr)
{
    struct MM4000_trans *trans = (struct MM4000_trans *) mr->dpvt;
    MOTOR_CALL *motor_call;
    char buff[20];
    int axis = mr->out.value.vmeio.signal + 1;
    double dval;

    buff[0] = '\0';

    motor_call = &(trans->motor_call);

    if (MM4000_table[command] > motor_call->type)
	motor_call->type = MM4000_table[command];

    if (trans->state != BUILD_STATE) return ERROR;

    switch (command)
    {
    case MOVE_ABS:
        dval = parms[0] * trans->drive_resolution + trans->offset;
        sprintf(buff, "%dPA%f;", axis, dval);
	break;
    case MOVE_REL:
        dval = parms[0] * trans->drive_resolution;
        sprintf(buff, "%dPR%f;", axis, dval);
	break;
    case HOME_FOR:
    case HOME_REV:
    case HOME_ENC:
        sprintf(buff, "%dOR;", axis);
	break;
    case LOAD_POS:
        /* The MM4000 can't be told where it is.  Keep track via an offset */
        trans->offset = trans->motor_pos - parms[0]*trans->drive_resolution;
        break;
    case SET_VEL_BASE:
        break;          /* MM4000 does not use base velocity */
    case SET_VELOCITY:
        dval = parms[0] * trans->drive_resolution;
        sprintf(buff, "%dVA%f;", axis, dval);
	break;
    case SET_ACCEL:
        /* 
         * The value passed is in steps/sec/sec.  
         * Convert to user units/sec/sec
         */
        dval = parms[0] * trans->drive_resolution;
        sprintf(buff, "%dAC%f;", axis, dval);
	break;
    case GO:
        /* 
         * The MM4000 starts moving immediately on move commands, GO command
         * does nothing
         */
	break;
    case SET_ENC_RATIO:
        /*
         * The MM4000 does not have the concept of encoder ratio, ignore this
         * command
         */
	break;
    case GET_MOTOR_POS:
    case GET_ENCODER_POS:
    case GET_INFO:
        /* These commands are not actually done by sending a message, but
           rather they will indirectly cause the driver to read the status
           of all motors */
	break;
    case STOP_AXIS:
        sprintf(buff, "%dST;", axis);
	break;
    case JOG:
        /* 
        * MM4000 does not have a jog command.  Simulate with move absolute
        * to the appropriate software limit. We can move to MM4000 soft limits.
        * If the record soft limits are set tighter than the MM4000 limits
        * the record will prevent JOG motion beyond its soft limits 
        */
        dval = parms[0] * trans->drive_resolution;
        sprintf(buff, "%dVA%f;", axis, ABS(dval));
        strcat(motor_call->message, buff);
        if (dval > 0.) {
            /* This is a positive move in MM4000 coordinates */
            sprintf(buff, "%dPA%f;", axis, trans->high_limit);
        } else {
            /* This is a negative move in MM4000 coordinates */
            sprintf(buff, "%dPA%f;", axis, trans->low_limit);
        }
	break;
    }
    strcat(motor_call->message, buff);

    return (0);
}


/* callback from MM4000 driver for motor in motion */
STATIC void MM4000_callback(MOTOR_RETURN * motor_return)
{
    struct motorRecord *mr = (struct motorRecord *) motor_return->precord;
    struct rset *prset = (struct rset *) (motor_return->precord->rset);
    struct MM4000_trans *ptrans;

    Debug(2, "MM4000_callback: entry\n", 0);
    ptrans = (struct MM4000_trans *) mr->dpvt;

    Debug(6, "MM4000_callback: FASTLOCK\n", 0);
    FASTLOCK(&ptrans->lock);

    Debug(6, "MM4000_callback: update ptrans\n", 0);
    ptrans->callback_changed = YES;
    /* raw motor pulse count */
    ptrans->motor_pos = motor_return->position;
    Debug(6, "MM4000_callback: pos=%f\n", ptrans->motor_pos);
    /* raw encoder pulse count */
    ptrans->encoder_pos = motor_return->encoder_position;
    Debug(6, "MM4000_callback: enc. pos=%f\n", ptrans->encoder_pos);
    /* raw motor velocity */
    ptrans->vel = motor_return->velocity;
    Debug(6, "MM4000_callback: vel=%f\n", ptrans->vel);

    ptrans->status = motor_return->status;	/* status */
    Debug(6, "MM4000_callback: status=%x\n", ptrans->status);

    Debug(6, "MM4000_callback: FASTUNLOCK\n", 0);
    FASTUNLOCK(&ptrans->lock);

    /* free the return data buffer */
    Debug(6, "MM4000_callback: free MOTOR_RETURN\n", 0);
    (*MM4000_access.free) (motor_return);

    Debug(6, "MM4000_callback: dbScanLock\n", 0);
    dbScanLock((struct dbCommon *) mr);
    Debug(6, "MM4000_callback: calling process\n", 0);
    (*prset->process) ((struct dbCommon *) mr);
    Debug(6, "MM4000_callback: dbScanUnlock\n", 0);
    dbScanUnlock((struct dbCommon *) mr);
    Debug(2, "MM4000_callback: exit..\n", 0);
    return;
}
